! This file is part of FD4: Four-Dimensional Distributed Dynamic Data structures
!  
! Copyright 2010 Matthias Lieber, ZIH, TU Dresden, Germany
! matthias.lieber@tu-dresden.de
!  
! FD4 is free software: you can redistribute it and/or modify
! it under the terms of the GNU General Public License as published by
! the Free Software Foundation, either version 3 of the License, or
! (at your option) any later version.
!  
! FD4 is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
! GNU General Public License for more details.
!  
! You should have received a copy of the GNU General Public License
! along with FD4.  If not, see <http://www.gnu.org/licenses/>.

!--------------------------------------------------------------!
!                                                              !
!                  Boundary condition handling                 !
!                                                              !
!--------------------------------------------------------------!

#include "fd4.h"

!! Apply boundary conditions on a domain.
!!
!! Currently we have zero gradient and a simple method to set specific values.
!!
!! @see fd4_domain_mod
!! @author Matthias Lieber
module fd4_boundary_mod

  use fd4_domain_mod
  use fd4_block_mod
  use fd4_globaldef_mod
  use fd4_vartab_mod
  use fd4_ghostlist_mod
  
  implicit none
  private
  
  public :: fd4_boundary_zerograd
  public :: fd4_boundary_zerograd_block
  public :: fd4_boundary_spec

contains

  !! Apply zero gradient boundary conditions on all blocks of a domain.
  !!
  !! Sets the given cell-centered vars in the boundary ghost block to the same value as
  !! in the cells at domain boundary.
  !! Sets the value of the given face variables at domain boundary (within domain) to vnull.
  !!
  !! *TODO*:
  !! - ugh, dynamic vars?
  !!
  subroutine fd4_boundary_zerograd(domain, idx, st, nfields)
    implicit none
    type(fd4_domain), intent(inout) :: domain        !! the domain
    integer(i_k)    , intent(in)    :: nfields       !! size of $idx amd $st
    integer(i_k)    , intent(in)    :: idx(nfields)  !! array of field indices
    integer(i_k)    , intent(in)    :: st(nfields)   !! array of time step indices
    
    type(fd4_block) , pointer       :: bgb    ! the boundary ghost block
    type(fd4_block) , pointer       :: b      ! the (ordinary) block at domain boundary

#if ( FD4_VERBOSE_LEVEL >= 3 )
      integer(i_k) :: f
      ! error checks
      do f=1,nfields
        if(idx(f)<0 .or. idx(f)>ubound(domain%vartab,1)) then
          FD4_L1_ERROR('field idx not valid f='//i2s(f))
        end if
        if(st(f)<0 .or. st(f)>domain%vartab(idx(f))%nsteps) then
          FD4_L1_ERROR('time step not valid f='//i2s(f))
        end if
      end do
#endif

    ! will be null if we have periodic bounds or no blocks at boundary
    bgb => domain%glist(FD4_GHOST_BND)%start
    
    ! loop over all boundary ghost blocks
    do while(associated(bgb))
      
      b => bgb%neigh(2,1)%l  ! the boundary block within the domain
      
      call fd4_boundary_zerograd_set(domain, b, bgb, idx, st, nfields)
    
      bgb => bgb%neigh(1,2)%l  ! iterate to next
    end do
  
  end subroutine fd4_boundary_zerograd




  !! Apply zero gradient boundary conditions on specified block
  !!
  !! Sets the given cell-centered vars in the boundary ghost block to the same value as
  !! in the cells at domain boundary.
  !! Sets the value of the given face variables at domain boundary (within domain) to vnull.
  !!
  !! *TODO*:
  !! - ugh, dynamic vars?
  !!
  subroutine fd4_boundary_zerograd_block(domain, b, idx, st, nfields)
    implicit none
    type(fd4_domain), intent(inout) :: domain        !! the domain
    type(fd4_block) , pointer       :: b             !! the block
    integer(i_k)    , intent(in)    :: nfields       !! size of $idx amd $st
    integer(i_k)    , intent(in)    :: idx(nfields)  !! array of field indices
    integer(i_k)    , intent(in)    :: st(nfields)   !! array of time step indices
    
    integer(i_k) :: dim

#if ( FD4_VERBOSE_LEVEL >= 3 )
      integer(i_k) :: f
      ! error checks
      do f=1,nfields
        if(idx(f)<0 .or. idx(f)>ubound(domain%vartab,1)) then
          FD4_L1_ERROR('field idx not valid f='//i2s(f))
        end if
        if(st(f)<0 .or. st(f)>domain%vartab(idx(f))%nsteps) then
          FD4_L1_ERROR('time step not valid f='//i2s(f))
        end if
      end do
#endif

    ! will be null if we have periodic bounds or no blocks at boundary
    if(.not. associated(domain%glist(FD4_GHOST_BND)%start)) return
    
    ! identify boundary ghost blocks
    do dim=1,3
      if(domain%periodic(dim)) cycle
      
      ! lower boundary
      if(b%pos(dim)==1) then
        call fd4_boundary_zerograd_set(domain, b, b%neigh(dim,1)%l, idx, st, nfields)
      end if
      
      ! upper boundary
      if(b%pos(dim)==domain%blnum(dim)) then
        call fd4_boundary_zerograd_set(domain, b, b%neigh(dim,2)%l, idx, st, nfields)
      end if

    end do
    
  end subroutine fd4_boundary_zerograd_block


  
  !! Apply zero gradient boundary conditions on specified block/boundary ghost block pair.
  !!
  !! Internal subroutine. Used by $fd4_boundary_zerograd and $fd4_boundary_zerograd_block.
  subroutine fd4_boundary_zerograd_set(domain, b, bgb, idx, st, nfields)
    implicit none
    type(fd4_domain), intent(inout) :: domain        !! the domain
    type(fd4_block) , pointer       :: b             !! the block
    type(fd4_block) , pointer       :: bgb           !! the boundary ghost block
    integer(i_k)    , intent(in)    :: nfields       !! size of $idx amd $st
    integer(i_k)    , intent(in)    :: idx(nfields)  !! array of field indices
    integer(i_k)    , intent(in)    :: st(nfields)   !! array of time step indices
    
    integer(i_k) :: sc, sf, g, f

    ! bgb%pos(3) = boundary ghost block indicator (must be 1)
    ! bgb%pos(3) = boundary ghost block dimension relative to normal block (x,y,z resp. 1,2,3)
    ! bgb%pos(2) = boundary ghost block direction relative to normal block ( -1,1 resp. 1,2)
    
    if(bgb%pos(2)==1) then
      sc = 1                 ! bgb is before b, lower slice is copied   (cell centered)
      sf = 1                 ! bgb is before b, lower slice is set to 0 (facevar)
    else
      sc = b%ext(bgb%pos(3)) ! bgb is behind b, upper slice is copied   (cell centered)
      sf = sc + 1            ! bgb is behind b, upper slice is set to 0 (facevar)
    end if

    ! x
    if(bgb%pos(3)==1) then      
      do f=1,nfields
        if( domain%vartab(idx(f))%facevar == 1 ) then
          b%fields(idx(f),st(f))%l(:,sf,:,:) = domain%vartab(idx(f))%vnull
        else if( domain%vartab(idx(f))%facevar == 0 ) then        
          do g=1,domain%nghosts(1)
            bgb%fields(idx(f),st(f))%l(:,g,:,:) = b%fields(idx(f),st(f))%l(:,sc,:,:)
          end do
        end if
      end do
    end if

    ! y
    if(bgb%pos(3)==2) then      
      do f=1,nfields
        if( domain%vartab(idx(f))%facevar == 2 ) then
          b%fields(idx(f),st(f))%l(:,:,sf,:) = domain%vartab(idx(f))%vnull
        else if( domain%vartab(idx(f))%facevar == 0 ) then
          do g=1,domain%nghosts(2)
            bgb%fields(idx(f),st(f))%l(:,:,g,:) = b%fields(idx(f),st(f))%l(:,:,sc,:)
          end do
        end if
      end do
    end if

    ! z
    if(bgb%pos(3)==3) then      
      do f=1,nfields
        if( domain%vartab(idx(f))%facevar == 3 ) then
          b%fields(idx(f),st(f))%l(:,:,:,sf) = domain%vartab(idx(f))%vnull
        else if( domain%vartab(idx(f))%facevar == 0 ) then
          do g=1,domain%nghosts(3)
            bgb%fields(idx(f),st(f))%l(:,:,:,g) = b%fields(idx(f),st(f))%l(:,:,:,sc)
          end do
        end if
      end do
    end if
  
  end subroutine fd4_boundary_zerograd_set






  !! Apply specific gradient boundary conditions on all blocks of a domain.
  !!
  !! $value(1) is used for the first boundary layer, $value(2) for the second etc.
  !! Does only process cell centered variables. Only one spatial dimension is processed.
  !!
  subroutine fd4_boundary_spec(domain, idx, st, nfields, face, lvalues, uvalues)
    implicit none
    type(fd4_domain), intent(inout) :: domain        !! the domain
    integer(i_k)    , intent(in)    :: nfields       !! size of $idx amd $st
    integer(i_k)    , intent(in)    :: idx(nfields)  !! array of field indices
    integer(i_k)    , intent(in)    :: st(nfields)   !! array of time step indices
    integer(i_k)    , intent(in)    :: face          !! dimension to process (1,2,3)
    real(r_k)       , intent(in)    :: lvalues(:)    !! values for each ghost plane (lower boundary)
    real(r_k)       , intent(in)    :: uvalues(:)    !! values for each ghost plane (upper boundary)
    
    type(fd4_block) , pointer       :: bgb    ! the boundary ghost block

#if ( FD4_VERBOSE_LEVEL >= 3 )
      integer(i_k) :: f
      ! error checks
      do f=1,nfields
        if(idx(f)<0 .or. idx(f)>ubound(domain%vartab,1)) then
          FD4_L1_ERROR('field idx not valid f='//i2s(f))
        end if
        if(st(f)<0 .or. st(f)>domain%vartab(idx(f))%nsteps) then
          FD4_L1_ERROR('time step not valid f='//i2s(f))
        end if
        if(domain%vartab(idx(f))%facevar /= 0) then
          FD4_L1_ERROR('not a cell centered variable f='//i2s(f))
        end if
      end do
#endif

    ! will be null if we have periodic bounds or no blocks at boundary
    bgb => domain%glist(FD4_GHOST_BND)%start
    
    ! loop over all boundary ghost blocks
    do while(associated(bgb))
      
      if(bgb%pos(3)==face) then
        call fd4_boundary_spec_set(domain, bgb, idx, st, nfields, lvalues, uvalues)
      end if
    
      bgb => bgb%neigh(1,2)%l  ! iterate to next
    end do
  
  end subroutine fd4_boundary_spec






  !! Apply specific boundary conditions on specified block/boundary ghost block pair.
  !!
  !! Internal subroutine. Used by $fd4_boundary_spec.
  subroutine fd4_boundary_spec_set(domain, bgb, idx, st, nfields, lvalues, uvalues)
    implicit none
    type(fd4_domain), intent(inout) :: domain        !! the domain
    type(fd4_block) , pointer       :: bgb           !! the boundary ghost block
    integer(i_k)    , intent(in)    :: nfields       !! size of $idx amd $st
    integer(i_k)    , intent(in)    :: idx(nfields)  !! array of field indices
    integer(i_k)    , intent(in)    :: st(nfields)   !! array of time step indices
    real(r_k),target, intent(in)    :: lvalues(:)    !! values for each ghost plane (lower boundary)
    real(r_k),target, intent(in)    :: uvalues(:)    !! values for each ghost plane (upper boundary)
    
    integer(i_k) :: g, f
    real(r_k), pointer :: values(:)

    ! bgb%pos(3) = boundary ghost block indicator (must be 1)
    ! bgb%pos(3) = boundary ghost block dimension relative to normal block (x,y,z resp. 1,2,3)
    ! bgb%pos(2) = boundary ghost block direction relative to normal block ( -1,1 resp. 1,2)
    
    if(bgb%pos(2) == 1) then
      values => lvalues
    else
      values => uvalues
    end if
    
    ! x
    if(bgb%pos(3)==1) then      
      do f=1,nfields
        do g=1,domain%nghosts(1)
          bgb%fields(idx(f),st(f))%l(:,g,:,:) = values(g)
        end do
      end do
    end if

    ! y
    if(bgb%pos(3)==2) then      
      do f=1,nfields
        do g=1,domain%nghosts(2)
          bgb%fields(idx(f),st(f))%l(:,:,g,:) = values(g)
        end do
      end do
    end if

    ! z
    if(bgb%pos(3)==3) then      
      do f=1,nfields
        do g=1,domain%nghosts(3)
          bgb%fields(idx(f),st(f))%l(:,:,:,g) = values(g)
        end do
      end do
    end if
  
  end subroutine fd4_boundary_spec_set




end module fd4_boundary_mod
